 
 
local dbg_stuff = false

local layer_digits = 1
local layer_decimal = 2

local layer_guide = 3
local layer_guidebox = 4

local layer_train = 5
local layer_puffs = 6

local layer_grenade = 7

local layer_laser = 7.5
local layer_turbolaser = 7.6

local layer_cannon = 8
local layer_turbocannon = 9
local layer_track = 10

local layer_footballer = 11

local layer_artillery = 12
local layer_artshot = 13

local layer_crosshair = 14

local layer_klok = 17

local layer_kaboom = 18


local _G = _G

function makegame()

local world = setmetatable({}, {__index = _G})
world._G = world
setfenv(1, world)


local tixers = {}

local rhythms = {}
local rhythm_frame = 0
local death_frame = 30 * 60
function AddRhythm(n, d, ofs, func)
  table.insert(rhythms, {n = n, d = d, ofs = ofs, func = func})
end
function IsRhythm(n, d, ofs)
  local ticker = 60 * n / d
  return math.mod(rhythm_frame - ofs, ticker) == 0
end

do
  local klok = CreateFrame("Sprite")
  klok:SetLayer(layer_klok)
  klok:SetTexture("clock")
  klok:SetPoint("TOPRIGHT", UIParent, "TOPRIGHT", -10, 10)
  
  local klok_hands = CreateFrame("Sprite", klok)
  klok_hands:SetAllPoints()
  function klok_hands:Draw()
    local sx, sy, ex, ey = self:GetBounds()
    local mx, my = (sx + ex) / 2, (sy + ey) / 2
    local daxis = mx - sx
    
    gl.LineWidth(1)
    gl.Enable("LINE_SMOOTH")
  
    gl.Color(0.4, 0.4, 0.3)
    
    gl.Begin("LINES")
    for i = 1, 30 do
      gl.Vertex(mx + math.sin(i / 30 * 3.14159 * 2) * daxis * 0.60, my + math.cos(i / 30 * 3.14159 * 2) * daxis * 0.65)
      gl.Vertex(mx + math.sin(i / 30 * 3.14159 * 2) * daxis * 0.70, my + math.cos(i / 30 * 3.14159 * 2) * daxis * 0.65)
    end
    gl.End()
    
    gl.LineWidth(3)
    
    local death_second = math.floor(death_frame / 60)
    local rhythm_second = math.floor(rhythm_frame / 60)
    if death_second >= rhythm_second + 30 then death_second = rhythm_second end
    
    death_second = math.mod(death_second, 30) - 7.5
    rhythm_second = math.mod(rhythm_second, 30) - 7.5
    
    gl.Color(0.8, 0.4, 0.4)
    gl.Begin("LINES")
    gl.Vertex(mx, my)
    gl.Vertex(mx + math.cos(death_second / 30 * 3.14159 * 2) * daxis * 0.7, my + math.sin(death_second / 30 * 3.14159 * 2) * daxis * 0.7)
    
    gl.Color(0.7, 0.7, 0.6)
    gl.Vertex(mx, my)
    gl.Vertex(mx + math.cos(rhythm_second / 30 * 3.14159 * 2) * daxis * 0.8, my + math.sin(rhythm_second / 30 * 3.14159 * 2) * daxis * 0.8)
    gl.End()
  end
  
  AddRhythm(1, 1, 0, function () PlaySound("tick") end)
end

--[[
*************************************************
* digits
*
]]

local digis = {}
for k = 0, 9 do
  digis[k] = Texture("digit_" .. k)
end

local digit_queue = {}
local digit_queue_reverse = {}
local digit_anchor = CreateFrame("Text")
digit_anchor:SetLayer(layer_decimal)
digit_anchor:SetText(".")
digit_anchor:SetColor(1, 1, 1)
digit_anchor:SetSize(61)
digit_anchor:SetPoint("CENTER", UIParent, "CENTER", 140, 0)
digit_queue[0] = digit_anchor

local increment_cooldown = 0

local queuecrement = nil

local function find_digit(dig)
  local diggit, side
  for k, v in ipairs(digit_queue) do
    if v == dig then
      diggit, side = k, digit_queue
    end
  end
  for k, v in ipairs(digit_queue_reverse) do
    if v == dig then
      diggit, side = k, digit_queue_reverse
    end
  end
  assert(diggit)
  return diggit, side
end

local function add_digit()
  local ndig = MakeDigit()
  ndig:SetConnector(digit_queue[#digit_queue])
  table.insert(digit_queue, ndig)
  return ndig
end
local commas = {}
local function standardize_digits()
  -- strip leading 0's
  while #digit_queue > 1 and digit_queue[#digit_queue].current_digit == 0 do
    if queuecrement == digit_queue[#digit_queue] then queuecrement = nil end
    digit_queue[#digit_queue]:Detach()
    tixers[digit_queue[#digit_queue]] = nil
    digit_queue[#digit_queue] = nil
  end
  -- strip trailing 0's
  while #digit_queue_reverse > 0 and digit_queue_reverse[1].current_digit == 0 do
    if queuecrement == digit_queue_reverse[1] then queuecrement = nil end
    digit_queue_reverse[1]:Detach()
    tixers[digit_queue_reverse[1]] = nil
    table.remove(digit_queue_reverse, 1)
  end

  -- add a digit if needed
  if #digit_queue == 0 then add_digit():Transition(0) end
  
  -- commas
  for _, v in pairs(commas) do
    v:Hide()
  end
  for i = 4, #digit_queue, 3 do
    if not commas[i] then
      commas[i] = CreateFrame("Text")
      commas[i]:SetLayer(layer_decimal)
      commas[i]:SetText(",")
      commas[i]:SetSize(61)
      commas[i]:SetColor(1, 1, 1)
    end
    commas[i]:SetPoint("TOPLEFT", digit_queue[i], "TOPRIGHT", -4, 4)
    commas[i]:Show()
  end
  
  if #digit_queue_reverse > 0 then
    digit_anchor:SetColor(1, 1, 1)
  else
    digit_anchor:SetColor(1, 1, 1)
  end
end
local function increment_digit(dig, first)
  if increment_cooldown > rhythm_frame then
    if increment_cooldown <= rhythm_frame + 18 then
      queuecrement = dig
    end
    return
  end
    
  local diggit, side = find_digit(dig)
  
  if diggit == 1 and dig.current_digit == 3 and #digit_queue == 2 then
    text_event("tenbux")
  end
  if diggit == 2 and first then text_event("digittwo") end
  
  if dig.current_digit == 9 then
    dig:Transition(0, 60)
        
    if side[diggit + 1] then
      increment_digit(side[diggit + 1])
    elseif side == digit_queue_reverse then
      increment_digit(digit_queue[1])
    else
      add_digit():Transition(1, 60)
    end
  else
    dig:Transition(dig.current_digit + 1, 60)
  end
  
  increment_cooldown = rhythm_frame + 60
  
  standardize_digits()
  
  if first then PlaySound("number_shift") end
end
AddRhythm(1, 60, 0, function ()
  if queuecrement then
    local qc = queuecrement
    queuecrement = nil
    increment_digit(qc, true) -- might just reset it
  end
end)
local function obliterate_digit(dig)
  if queuecrement == dig then queuecrement = nil end
  
  local diggit, side = find_digit(dig)
  
  print("oblit")
  if side == digit_queue and diggit ~= #digit_queue then
    print("oblit normal")
    side[diggit + 1]:SetConnector(side[diggit - 1])
  elseif side == digit_queue_reverse and diggit ~= #digit_queue_reverse and diggit ~= 1 then
    print("oblit reverse", diggit, #digit_queue_reverse)
    side[diggit - 1]:SetConnector(side[diggit + 1], true)
  elseif side == digit_queue_reverse and diggit == #digit_queue_reverse and side[diggit - 1] then
    print("oblit reverse end")
    side[diggit - 1]:SetConnector(digit_anchor, true)
  end
  print("oblit done")
  
  table.remove(side, diggit)  -- we really want a smooth fall effect here, buuuut
  tixers[dig] = nil
  dig:Detach() -- go 'way
  
  standardize_digits()
  
  PlaySound("number_shatter", 0.5)
end
local function decrement_digit(dig, first)
  assert(dig)
  --print("decrementing", dig, dig.current_digit)
  
  local diggit, side = find_digit(dig)

  if dig.current_digit == 0 then
    dig:Transition(9, 20, true)
    
    local diggit, side = find_digit(dig)
        
    if side[diggit + 1] then
      decrement_digit(digit_queue[diggit + 1])
    elseif side == digit_queue_reverse then
      decrement_digit(digit_queue[1])
    else
      -- decrementing 1 to 0 somehow
    end
  else
    dig:Transition(dig.current_digit - 1, 20, true)
  end
  
  if first then PlaySound("number_rollback") end
  standardize_digits()
end

do
  local function RenderDigit(digit, sx, sy, ex, ey, ofs)
    local dy = ey - sy
    
    local tsy, tey = 0, 1
    
    if ofs <= -1 or ofs >= 1 then return end
    
    if ofs < 0 then
      tsy = -ofs
    end
    
    if ofs > 0 then
      tey = 1 - ofs
    end
    
    local nsy, ney = sy + dy * ofs, ey + dy * ofs
    nsy = math.max(sy, math.min(ey, nsy))
    ney = math.max(sy, math.min(ey, ney))
    
    glutil.RenderBoundedSpriteSegment(digit, sx, nsy, ex, ney, 0, tsy, 1, tey)
  end
  local function dofs(tix)
    return tix - 1
  end
  
  local function get_intended_location(self)
    return self.connector:GetXCenter() - 28 * self.direction + (28 - self.connector:GetWidth()) / 2 * self.direction
  end
  
  local digit_proto = {}
  function digit_proto:Draw()
    local ofs = dofs(self.transition_position and (self.transition_position / self.transition_frames) or 1)
    
    local sx, sy, ex, ey = self:GetBounds()
    local dy = ey - sy
    
    if self.transition_reverse then ofs = -ofs end
    
    if self.last_digit then
      RenderDigit(digis[self.last_digit], sx, sy, ex, ey, ofs + (self.transition_reverse and -1 or 1))
    end
    
    RenderDigit(digis[self.current_digit], sx, sy, ex, ey, ofs)
  end
  function digit_proto:Transition(newdigit, framedelay, reverse)
    if framedelay then
      self.last_digit = self.current_digit
      self.current_digit = newdigit
      self.transition_position = 0
      self.transition_frames = framedelay
      self.transition_reverse = reverse
      self:Tick()
    else
      self.current_digit = newdigit
      self.transition_position, self.transition_frames, self.transition_reverse = nil, nil, nil
    end
  end
  function digit_proto:Tick()
    if self.transition_position then
      self.transition_position = self.transition_position + 1
      
      if self.transition_position == self.transition_frames then
        self:Transition(self.current_digit)
      end
    end
    
    local intended_location = get_intended_location(self)
    local current_location = self:GetXCenter()
    
    if intended_location ~= current_location then
      if not (self.direction == 1 and current_location > intended_location) then
        local newloc = approach(current_location, intended_location, 0.5)
        self:SetPoint(0.5, nil, UIParent, "LEFT", newloc, nil)
      end
    end
  end
  function digit_proto:Key(button, ascii, event)
    if event ~= "press" then return end
    
    if button == "mouse_left" then
      increment_digit(self, true)
    end
  end
  function digit_proto:SetConnector(link, backwards)
    local flat = false
    if not self.connector then
      flat = true
    end
    self.connector = link
    self.direction = backwards and -1 or 1
    if flat then
      self:SetPoint("TOP", digit_anchor, "TOP")
      self:SetPoint(0.5, nil, UIParent, "LEFT", get_intended_location(self), nil)
      print("setcflat", self.connector, self.direction, get_intended_location(self), self.connector:GetXCenter(), self.direction, self.connector:GetWidth())
      --self.connector:GetXCenter() - 28 * self.direction + (28 - self.connector:GetWidth()) / 2 * self.direction
    end
  end
  
  digit_proto.Click = increment_digit
  function MakeDigit()
    local nd = CreateFrame("Frame")
    nd:SetLayer(layer_digits)
    nd:SetWidth(28)
    nd:SetHeight(46)
    
    for k, v in pairs(digit_proto) do
      nd[k] = v
    end
    
    tixers[nd] = true
    
    return nd
  end
end

if dbg_stuff then
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
  add_digit():Transition(9)
else
  add_digit():Transition(0)
end

local digit_anchor_left = CreateFrame("Frame")
digit_anchor_left:SetWidth(digit_queue[1]:GetWidth())
digit_anchor_left:SetPoint("TOPRIGHT", digit_anchor, "TOPLEFT")

local digit_anchor_right = CreateFrame("Frame")
digit_anchor_right:SetWidth(digit_queue[1]:GetWidth())
digit_anchor_right:SetPoint("TOPLEFT", digit_anchor, "TOPRIGHT")



--[[
*************************************************
* actual game logic
*
]]

if dbg_stuff then
  grenade_start = 1000
  train_start = 1111
  laser_start = 700
  football_start = 1
  hsgrenade_start = 100
  turbolaser_start = 100
  artillery_start = 13
  bigartillery_start = 12
else
  grenade_start = 4
  train_start = 6
  laser_start = 7
  football_start = 9
  hsgrenade_start = 10
  turbolaser_start = 11
  artillery_start = 13
  bigartillery_start = 14
end

--[[
*************************************************
* grenade (4+)
*
]]

local function kaboom(x, y)
  local boom = CreateFrame("Sprite")
  boom:SetTexture("kaboom_1")
  boom:SetLayer(layer_kaboom)
  tixers[boom] = true
  boom:SetPoint("CENTER", UIParent, "TOPLEFT", x, y)
  boom.Tick = coroutine.wrap(function()
    coroutine.pause(7)
    boom:SetTexture("kaboom_2")
    coroutine.pause(8)
    boom:SetTexture("kaboom_3")
    coroutine.pause(7)
    boom:SetTexture("kaboom_4")
    coroutine.pause(8)
    tixers[boom] = nil
    boom:Detach()
  end)
end

local function ChuckGrenade(decimally)
  text_event("grenade")
  
  PlaySound("grenade_toss")
  local gren = CreateFrame("Sprite")
  gren:SetLayer(layer_grenade)
  gren:SetTexture("grenade")
  tixers[gren] = true
  
  local total_life = 180
  local start_vel = 4
  local gravity = (start_vel * 2) / total_life
  
  gren.x = 1024
  gren.y = digit_anchor:GetYCenter()
  gren.yv = start_vel
  gren.lifetime = total_life
  
  gren.Tick = coroutine.wrap(function()
    while gren.lifetime > 0 do
      local targx = (digit_queue_reverse[1] and decimally) and digit_queue_reverse[1]:GetXCenter() or digit_queue[1]:GetXCenter()
      gren.x = ((gren.x * (gren.lifetime - 1)) + targx) / gren.lifetime
      gren.y = gren.y - gren.yv
      gren.yv = gren.yv - gravity
      gren.lifetime = gren.lifetime - 1
      
      gren:SetPoint("CENTER", UIParent, "TOPLEFT", gren.x, gren.y)
      
      coroutine.yield()
    end
    
    -- kaboom
    text_event("grenade_fail")
    obliterate_digit((digit_queue_reverse[1] and decimally) and digit_queue_reverse[1] or digit_queue[1])
    gren:Detach()
    tixers[gren] = nil
    PlaySound("grenade_boom")
    kaboom(gren.x, gren.y)
  end)
  function gren:Key(button, ascii, event)
    if event == "press" and button == "mouse_left" then
      text_event("grenade_success")
      gren:Detach()
      tixers[gren] = nil
      PlaySound("grenade_boom")
      kaboom(gren.x, gren.y)
    end
  end
  gren:Tick()
end


local add_grenade = false
AddRhythm(3, 1, 0, function () 
  if add_grenade then
    ChuckGrenade()
  end
  add_grenade = false
end)

-- grenades. oh shut the fuck up >:(
local orig_add_dig = add_digit
function add_digit()
  local rv = orig_add_dig()
  if #digit_queue >= grenade_start then
    add_grenade = true
  end
  return rv
end

--[[
*************************************************
* train (6+)
*
]]

local tofs = 0
AddRhythm(3, 1, 90, function ()
  if #digit_queue >= train_start then
    --PlaySound("train_whistle")
    text_event("train")
    
    local tran_runner = CreateFrame("Frame")
    tran_runner:SetLayer(layer_train)
    
    local tran = CreateFrame("Sprite", tran_runner)
    tran:SetTexture("train")
    tixers[tran] = true
    
    tran:SetPoint("BOTTOM", UIParent, "BOTTOM") -- grunch
    tran:SetPoint("RIGHT", tran_runner, "RIGHT", 216 - 192)
    
    local thisofs = tofs
    tofs = tofs + 1
    
    local sx = -200
    local speed = digit_queue[1]:GetWidth() / 7.5
    local ctx = math.floor((digit_anchor_left:GetXCenter() - sx) / digit_queue[1]:GetWidth())
    
    function tran:Tick()
      sx = sx + speed
      tran_runner:SetPoint("RIGHT", UIParent, "LEFT", sx, 0)
      local ttx = math.floor((digit_anchor_left:GetXCenter() - sx) / digit_queue[1]:GetWidth())
      if ttx ~= ctx and math.mod(ttx + thisofs, 2) == 0 then
        local impact
        
        local idx = ttx + 2
        if idx > 0 and digit_queue[idx] then
          impact = true
          PlaySound("train_puff")
        else
          impact = false
        end
        
        local puff = CreateFrame("Sprite")
        puff:SetLayer(layer_puffs)
        puff:SetTexture("train_puff")
        if impact then puff:SetColor(197 / 255, 160 / 255, 251 / 255) end
        
        puff.x = sx
        puff.y = 768 - 20
        function puff:Tick()
          puff.y = puff.y - 3
          
          puff:SetPoint("CENTER", UIParent, "TOPLEFT", puff.x, puff.y)
          
          if not impact and puff.y < 700 then
            tixers[puff] = nil
            puff:Detach()
          end
          
          if impact and digit_queue[idx] and puff.y > digit_queue[idx]:GetTop() and puff.y < digit_queue[idx]:GetBottom() then
            decrement_digit(digit_queue[idx], true)
            tixers[puff] = nil
            puff:Detach()
          end
        end
        function puff:Key(button, ascii, event)
          if event == "press" and button == "mouse_left" then
            puff.Key = nil
            puff.Tick = coroutine.wrap(function ()
              puff:SetTexture("train_puff_disperse_1")
              coroutine.pause(10)
              puff:SetTexture("train_puff_disperse_2")
              coroutine.pause(10)
              puff:SetTexture("train_puff_disperse_3")
              coroutine.pause(10)
              puff:Detach()
              tixers[puff] = nil
            end)
          end
        end
        puff:Tick()

        tixers[puff] = true
        
        ctx = ttx
      end
      
      if sx > 1300 then
        tixers[tran] = nil
        tran_runner:Detach()
      end
    end
  end
end)

--[[
*************************************************
* imma firin' my catchphrase (7+)
*
]]

local track = CreateFrame("Frame")
track:SetLayer(layer_track)
local trackimg = Texture("laser_track")
function track:Draw()
  for i = 0, 1024, 64 do
    glutil.RenderBoundedSprite(trackimg, i, 0, i + 64, 32)
  end
end

local laser = CreateFrame("Sprite")
laser:SetLayer(layer_cannon)
laser:SetTexture("laser")
laser:SetPoint("TOP", UIParent, "TOP")
laser:SetPoint(0.5, nil, UIParent, "LEFT", -500, 0)
laser.x = -500

local function travel_to(self, target)
  coroutine.yield() -- just so we don't get a 2-second cycle
  
  local accel = 1
  local vdx = 0
  local vps = 0
  local dist = math.abs(target - self.x)
  local path = {}
  while vps < dist / 2 do
    vdx = vdx + accel
    vps = vps + vdx
    table.insert(path, vps)
  end
  --print("Path generated", #path)
  
  while not IsRhythm(1, 1, 60 - #path * 2) do coroutine.yield() end -- wait to synch
  
  if target ~= self.x then PlaySound("laser_shift") end
  
  local stfrom = self.x
  local mult = math.abs(target - self.x) / (target - self.x)
  
  for i = 1, #path do
    self.x = stfrom + path[i] * mult
    coroutine.yield()
  end
  
  for i = #path, 1, -1 do
    self.x = target - path[i] * mult
    coroutine.yield()
  end
  
  self.x = target
end
tixers[laser] = true

local function SpawnLaserBeam(x)
  local beam = CreateFrame("Frame")
  beam:SetLayer(layer_laser)
  local lifetotal = 30
  local lifeleft = lifetotal
  
  local lbeam = Texture("laser_beam")
  local lself = self
  function beam:Draw()
    for i = 0, 768, 32 do
      glutil.RenderBoundedSprite(lbeam, x - 16, i, x + 16, i + 32, 1, 1, 1, lifeleft / lifetotal)
    end
  end
  function beam:Tick()
    lifeleft = lifeleft - 1
    if lifeleft <= 0 then
      tixers[beam] = nil
      beam:Detach()
    end
  end
  tixers[beam] = true
end

local lcc
local function lasercoro(self)
  while true do
    if #digit_queue <= laser_start then
      travel_to(self, -500)
      coroutine.yield()
    else
      local pick = math.random(#digit_queue - 1)  -- bah
      travel_to(self, digit_queue[pick]:GetXCenter())
      
      -- shoot, 3 seconds
      text_event("laser")
      
      lcc = ControlSound("laser_charge")
      for i = 1, 12 do
        laser:SetTexture("laser_fire_1")
        coroutine.pause(7)
        laser:SetTexture("laser_fire_2")
        coroutine.pause(8)
      end
      laser:SetTexture("laser")
      
      if lcc then lcc:Stop() lcc = nil end
      
      if digit_queue[pick] then obliterate_digit(digit_queue[pick]) end
      PlaySound("laser_shot")
      SpawnLaserBeam(self.x)
    end
  end
end
local function smashcoro(self)
  self.frozen = true
  coroutine.pause((5 + math.random() * 4) * 60)
  self.coro = coroutine.wrap(lasercoro, self)
  self.frozen = false
end

laser.coro = coroutine.wrap(lasercoro, laser)
function laser:Tick()
  self.coro()
  laser:SetPoint(0.5, nil, UIParent, "LEFT", self.x)
end
function laser:Key(button, ascii, event)
  if event ~= "press" or button ~= "mouse_left" then return end
  
  if self.frozen then return end
  
  laser:SetTexture("laser")
  if lcc then lcc:Stop() lcc = nil end
  self.coro = coroutine.wrap(smashcoro, self)
end

--[[
*************************************************
* footballer (9+)
*
]]

AddRhythm(4, 1, 60, function()
  if #digit_queue >= football_start then
    local futbal = CreateFrame("Sprite")
    futbal:SetLayer(layer_footballer)
    
    futbal.x = -60
    futbal.y = 0
    
    function futbal:Tick()
      if not self.flying then futbal:SetTexture("football_" .. (math.mod(math.floor(rhythm_frame / 15), 2) + 1)) end
      futbal.x = futbal.x + 3
      
      if self.flying then
        self.y = self.y - self.dy
        self.dy = self.dy - 0.25
        
        if self.y >= 0 then
          self.y = 0
          self.flying = false
        end
      end
      
      if not self.flying then
        if IsRhythm(1, 4, 0) then
          PlaySound("footstep_" .. math.random(4))
        end
        
        local clix, cliq, clid = 100000, nil, nil
        for i = 1, #digit_queue do
          local dst = math.abs(self.x - digit_queue[i]:GetXCenter())
          if dst < clix then
            clix, cliq, clid = dst, digit_queue, i
          end
        end
        for i = 1, #digit_queue_reverse do
          local dst = math.abs(self.x - digit_queue_reverse[i]:GetXCenter())
          if dst < clix then
            clix, cliq, clid = dst, digit_queue_reverse, i
          end
        end
        
        if clix < 28 then
          PlaySound("football_shift")
          if cliq == digit_queue then
            -- staple the first item in post-period to the last item in pre-period
            verbositude = true
            if #digit_queue_reverse > 0 then digit_queue_reverse[#digit_queue_reverse]:SetConnector(digit_queue[1], true) end
            verbositude = false
            
            -- staple the last item in pre-period to the anchor
            digit_queue[1]:SetConnector(digit_anchor, true)
            
            -- move that item to post-period
            table.insert(digit_queue_reverse, table.remove(digit_queue, 1))
            
            -- staple the new last item to the anchor
            if digit_queue[1] then digit_queue[1]:SetConnector(digit_queue[0]) end
            
            -- jam in our new item
            local dij = MakeDigit()
            dij:Transition(0)
            table.insert(digit_queue, clid, dij)
            
            -- link next item to the new one, if there is a next
            if digit_queue[clid + 1] then digit_queue[clid + 1]:SetConnector(dij) end
            
            -- link current item to the last one
            dij:SetConnector(digit_queue[clid - 1])
          else
            clid = clid + 1 -- adjustment for ordering
            
            -- jam in our new item
            local dij = MakeDigit()
            dij:Transition(0)
            table.insert(digit_queue_reverse, clid, dij)
            
            -- link this item to the "next"
            if clid == #digit_queue_reverse then
              dij:SetConnector(digit_anchor, true)
            else
              dij:SetConnector(digit_queue_reverse[clid + 1], true)
            end
            
            -- link the previous item to this one
            if digit_queue_reverse[clid - 1] then
              digit_queue_reverse[clid - 1]:SetConnector(dij, true)
            end
          end
          
          standardize_digits()
          
          self:Detach()
          tixers[self] = nil
        end
      end
        
      
      self:SetPoint(0.5, nil, UIParent, "LEFT", futbal.x)
      self:SetPoint(nil, 0.5, digit_anchor_left, "CENTER", 0, futbal.y)
      
      if self.x > 1200 then
        self:Detach()
        tixers[self] = nil
      end
    end
    function futbal:Key(button, ascii, event)
      if not self.flying and button == "mouse_left" and event == "press" then
        PlaySound("football_bounce")
        self.flying = true
        self.dy = 10
      end
    end
    
    tixers[futbal] = true
  end
end)

--[[
*************************************************
* high-speed grenade (10+)
*
]]

AddRhythm(5, 1, 30, function () 
  if #digit_queue > hsgrenade_start then
    ChuckGrenade(true)
  end
end)

--[[
*************************************************
* turbolaser (11+)
*
]]

local tlaser = CreateFrame("Sprite")
tlaser:SetLayer(layer_turbocannon)
tlaser:SetTexture("laser_turbo")

local function tl_should_wake()
  return #digit_queue >= turbolaser_start and #digit_queue_reverse > 0
end

tlaser:SetPoint(0.5, nil, digit_anchor_right, 0.5, nil)
tlaser:SetPoint("TOP", UIParent, "TOP", 0, -80)

local tlcoro


local function tlretracto(delay)
  tlaser:SetTexture("laser_turbo")
  if tlaser:GetTop() == 0 then
    PlaySound("turbolaser_clunk")
    for i = 1, 6 do
      tlaser:SetPoint("TOP", UIParent, "TOP", 0, -(80 / 6 * i))
      coroutine.yield()
    end
  end
  
  if delay then coroutine.pause(delay) end
  
  tlaser.Tick = coroutine.wrap(tlcoro)
end

local tlc = nil

function tlcoro()
  while not (IsRhythm(1, 1, 0) and tl_should_wake()) do coroutine.yield() end
  -- clunk
  tlc = ControlSound("turbolaser_charge")
  PlaySound("turbolaser_clunk")
  for i = 1, 6 do
    tlaser:SetPoint("TOP", UIParent, "TOP", 0, -(80 / 6 * (6 - i)))
    coroutine.yield()
  end
  
  for i = 1, 4 do
    if not tl_should_wake() then tlretracto() end
    tlaser:SetTexture("laser_turbo_charge_1")
    coroutine.pause(i == 1 and 4 or 10)
    if not tl_should_wake() then tlretracto() end
    tlaser:SetTexture("laser_turbo_charge_2")
    coroutine.pause(10)
    if not tl_should_wake() then tlretracto() end
    tlaser:SetTexture("laser_turbo_charge_3")
    coroutine.pause(10)
    if not tl_should_wake() then tlretracto() end
  end
  
  if tlc then tlc:Stop() tlc = nil end
  
  while tl_should_wake() do
    PlaySound("turbolaser_shot")
    SpawnLaserBeam(tlaser:GetXCenter())
    obliterate_digit(digit_queue_reverse[#digit_queue_reverse])
    
    tlaser:SetTexture("laser_turbo_charge_1")
    coroutine.pause(5)
    tlaser:SetTexture("laser_turbo_charge_2")
    coroutine.pause(5)
    tlaser:SetTexture("laser_turbo_charge_3")
    coroutine.pause(5)
    tlaser:SetTexture("laser_turbo_charge_1")
    coroutine.pause(5)
    tlaser:SetTexture("laser_turbo_charge_2")
    coroutine.pause(5)
    tlaser:SetTexture("laser_turbo_charge_3")
    coroutine.pause(5)
  end
  
  tlretracto()
end

function tlaser:Key(button, ascii, event)
  if button == "mouse_left" and event == "press" then
    tlaser.Tick = coroutine.wrap(tlretracto, ((5 + math.random() * 3) * 60))
    if tlc then tlc:Stop() tlc = nil end
  end
end
tlaser.Tick = coroutine.wrap(tlcoro)

tixers[tlaser] = true

--[[
*************************************************
* artillery (13, 14)
*
]]


local art = CreateFrame("Sprite")
art:SetLayer(layer_artillery)
art:SetTexture("artillery")

art:SetPoint("TOPLEFT", UIParent, "TOPLEFT", -100, 500)

local crosshair = CreateFrame("Sprite")
crosshair:SetLayer(layer_crosshair)
crosshair:Hide()

local function art_setp(p)
  if p >= 1 then
    crosshair:SetPoint(0.5, 1, digit_anchor_left, 0.5, 0, -28 * (p - 1), -20)
  else
    crosshair:SetPoint(0.5, 1, digit_anchor_right, 0.5, 0, 28 * -p, -20)
  end
end

function artcoro()
  while true do
    print("artcoro")
    while not IsRhythm(1, 1, 0) do coroutine.yield() end
    
    local tgpt = #digit_queue
    
    local wid = 1
    if #digit_queue >= bigartillery_start then wid = 3 end
    
    crosshair:SetTexture("artillery_crosshair_" .. wid)
    art_setp(tgpt)
    art.clicked = nil
    
    PlaySound("artillery_lock")
    local delay = 7
    for i = 1, 8 * 60 do
      if art.clicked then tgpt = tgpt - 2  art_setp(tgpt) PlaySound("artillery_blip") end
      art.clicked = nil
      if tgpt > #digit_queue then tgpt = #digit_queue art_setp(tgpt) end
      crosshair:Show()
      coroutine.yield()
    end
    
    PlaySound("artillery_kaboomy")
    print("firing at", tgpt)
    
    local shot = CreateFrame("Sprite")
    shot:SetLayer(layer_artshot)
    shot:SetTexture("artillery_shell")
    tixers[shot] = true
    
    local total_life = 120
    local start_vel
    local gravity = 0.2
    local gravdown = 0
    local gravdown_dy = 0
    for i = 1, total_life do
      gravdown_dy = gravdown_dy + gravity
      gravdown = gravdown + gravdown_dy
    end
    
    shot.x = art:GetXCenter()
    shot.y = art:GetYCenter()
    shot:SetPoint("CENTER", UIParent, "TOPLEFT", shot.x, shot.y)
    
    local deltay = shot.y - digit_anchor_left:GetYCenter()
    start_vel = (gravdown + deltay) / total_life
    shot.yv = start_vel
    local dstx = crosshair:GetXCenter()
    
    local vx = (dstx - shot.x) / total_life
    
    shot.Tick = coroutine.wrap(function()
      for i = 1, total_life do
        shot.x = shot.x + vx
        shot.yv = shot.yv - gravity
        shot.y = shot.y - shot.yv
        shot:SetPoint("CENTER", UIParent, "TOPLEFT", shot.x, shot.y)
        coroutine.yield()
      end
      
      local targdigs = {}
      
      if wid == 3 then tgpt = tgpt + 1 end
      
      local function maybeinsert(x)
        if x then table.insert(targdigs, x) end
      end
      
      for k = 1, wid do
        if tgpt < 1 then
          maybeinsert(digit_queue_reverse[1 - tgpt])
        else
          maybeinsert(digit_queue[tgpt])
        end
        tgpt = tgpt - 1
      end
      
      kaboom(shot.x, shot.y)
      PlaySound("artillery_impact")
      for _, v in pairs(targdigs) do obliterate_digit(v) end
      
      shot:Detach()
      tixers[shot] = nil
    end)
    tixers[shot] = true
    
    coroutine.pause(60)
  end
end

function art:Tick()
  crosshair:Hide()
  if #digit_queue < artillery_start then
    art:SetPoint("LEFT", UIParent, "LEFT", math.max(-100, art:GetLeft() - 1))
    art.coro = nil
  else
    if art:GetLeft() == 50 then
      if not art.coro then art.coro = coroutine.wrap(artcoro) end
      art.coro()
    else
      art:SetPoint("LEFT", UIParent, "LEFT", math.min(50, art:GetLeft() + 1))
    end
  end
end
function art:Key(button, ascii, event)
  if button == "mouse_left" and (event == "press" or event == "press_double") then
    print("klikky")
    art.clicked = true
  end
end

tixers[art] = true



local dialogue = {
  intro = {
    "Heya there!",
    "Looks like you've got a number.",
    "That's not a very big number, though.",
    "Why don't ya try makin' the number bigger?",
    "{Click on the number} to make it bigger.",
    persist = true,
    phase = 1,
  },
  tenbux = {
    "Nice!",
    "That's not gonna make it big very fast, though.",
    "Try {clicking on the other digit}.",
    persist = true,
    phase = 2,
  },
  digittwo = {
    "Now you're gettin' someplace! <3",
    "If you don't break your record every {thirty seconds}, you lose.",
    "So keep making that number bigger!",
    phase = 3,
  },
  grenade = {
    "{LOOK OUT! GRENADE!}",
    phase = 4,
    instant = true,
  },
  grenade_fail = {
    "That wasn't good.",
    "Try {clicking the grenade} next time.",
    delay = 30,
    persist = true,
    phase = 5,
  },
  grenade_success = {
    "You got it!",
    "You're well on the way to {winning the game}.",
    "All you need is a score of {one quadrillion points}. That's a one, with {fifteen digits} after it.",
    "Shouldn't be too bad, right?",
    phase = 6,
  },
  train = {
    "Is that a . . . train?",
    phase = 7,
  },
  laser = {
    "Laser cannon! Be careful!",
    "This place is gettin' too dangerous! I'm outta here!",
    phase = 8,
    escape = true,
    instant = true,
    delay_late = 120,
  }
}

local last_text = nil

function text_event(event)
  if last_text and last_text.phase >= dialogue[event].phase then return end
  
  print("Starting event", event)
  last_text = dialogue[event]
end
text_event("intro")

local bgo_r, bgo_g, bgo_b = 0.5, 0.5, 1.0
local bgi_r, bgi_g, bgi_b = 0, 0, 0.4

local function drawtri(self, r, g, b)
  gl.Color(r, g, b)
  gl.Begin("TRIANGLES")
  local sx, sy, ex, ey = self:GetBounds()
  gl.Vertex(sx, sy)
  gl.Vertex(ex, sy)
  gl.Vertex(ex, ey)
  gl.End()
end

local helper = CreateFrame("Sprite")
helper:SetLayer(layer_guide)
tixers[helper] = true
helper:SetTexture("the_researcher_left")
helper:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 1200, 650)

local textbox = CreateFrame("Frame")
textbox:SetLayer(layer_guidebox)
textbox:SetPoint("BOTTOMRIGHT", helper, 0.5, 0, 0, -30)
textbox:SetWidth(500)
textbox:SetHeight(150)

local outertri = CreateFrame("Frame", textbox)
outertri:SetPoint("BOTTOMRIGHT", textbox, "BOTTOMRIGHT")
outertri:SetHeight(40)
outertri:SetWidth(40)
function outertri:Draw()
  drawtri(self, bgo_r, bgo_g, bgo_b)
end

local outerbox = CreateFrame("Frame", textbox)
outerbox:SetPoint("TOPLEFT", textbox, "TOPLEFT")
outerbox:SetPoint("RIGHT", textbox, "RIGHT")
outerbox:SetPoint("BOTTOM", outertri, "TOP")
outerbox:SetBackgroundColor(bgo_r, bgo_g, bgo_b)

local innerbox = CreateFrame("Frame", textbox)
innerbox:SetLayer(1)
innerbox:SetPoint("TOPLEFT", outerbox, "TOPLEFT", 10, 10)
innerbox:SetPoint("BOTTOMRIGHT", outerbox, "BOTTOMRIGHT", -10, -10)
innerbox:SetBackgroundColor(bgi_r, bgi_g, bgi_b)

local innertri = CreateFrame("Frame", textbox)
innertri:SetLayer(1)
innertri:SetPoint("BOTTOMRIGHT", outertri, "BOTTOMRIGHT", -10, -10 - 10 * math.sqrt(2))
innertri:SetPoint("TOP", innerbox, "BOTTOM")
innertri:SetWidth(innertri:GetHeight())
function innertri:Draw()
  drawtri(self, bgi_r, bgi_g, bgi_b)
end

local fahncy = CreateFrame("Text_Multiline", innerbox)
fahncy:SetPoint("TOPLEFT", innerbox, "TOPLEFT", 10, 10)
fahncy:SetPoint("BOTTOMRIGHT", innerbox, "BOTTOMRIGHT", 10, 10)
fahncy:SetText("hellos")

textbox:Hide()


helper.coro = coroutine.wrap(function()
  textbox:Hide()
  local xps = 1200
  while xps >= 800 do
    helper:SetPoint("LEFT", UIParent, "LEFT", xps, nil)
    xps = xps - 3
    if IsRhythm(1, 3, 0) then PlaySound("footstep_" .. math.random(4), 0.5) end
    coroutine.yield()
  end
  
  helper:SetTexture("the_researcher")
  
  coroutine.pause(60)
  
  while true do
    local ctext = last_text
    
    if last_text.delay then coroutine.pause(last_text.delay) end
    
    print("hmm")
    for _, line in ipairs(ctext) do
      local lin = ""
      for word in line:gmatch("[^ ]+") do
        print("wordin'", word)
        if #lin > 0 then lin = lin .. " " end
        lin = lin .. word
        fahncy:SetText(lin:gsub("{", "\1Cff8080\1"):gsub("}", "\1Cffffff\1"))
        textbox:Show()
        
        PlaySound("peanuts_" .. math.random(7)) 
        if not last_text.instant then for t = 1, (#word + 2) * 2 do if last_text ~= ctext then break end coroutine.yield() end end
        if last_text ~= ctext then break end
      end
      
      for t = 1, 120 do if last_text ~= ctext then break end coroutine.yield() end
      if last_text ~= ctext then break end
    end
    
    if last_text.delay_late then coroutine.pause(last_text.delay_late) end
    
    if not ctext.persist then textbox:Hide() end
    
    if last_text.escape then break end
    
    while last_text == ctext do coroutine.yield() end
  end
  
  helper:SetTexture("the_researcher_flee")
  
  while xps <= 1200 do
    helper:SetPoint("LEFT", UIParent, "LEFT", xps, nil)
    xps = xps + 5
    if IsRhythm(1, 4, 0) then PlaySound("footstep_" .. math.random(4), 0.5) end
    coroutine.yield()
  end
  
  while true do coroutine.yield() end
end)
function helper:Tick()
  helper.coro()
end
  
--[[
*************************************************
* interface
*
]]

local best_val = "0"

function tick_loop()
  rhythm_frame = rhythm_frame + 1
  for _, v in pairs(rhythms) do
    if IsRhythm(v.n, v.d, v.ofs) then v.func() end
  end
  
  for k in pairs(tixers) do
    k:Tick()
  end
  
  local cval = ""
  
  for _, v in ipairs(digit_queue) do
    cval = v.current_digit .. cval
  end
  
  while #cval < #best_val do cval = "0" .. cval end
  while #best_val < #cval do best_val = "0" .. best_val end
  
  if best_val < cval then
    print("New record", cval)
    best_val = cval
    death_frame = rhythm_frame + 30*60
  end
  
  if #best_val >= 15 and best_val >= "1000000000000000" then PlaySound("gong") return {"win", rhythm_frame} end
  if death_frame <= rhythm_frame then PlaySound("gong") return {"lose", best_val} end
end
function render(...)
  UIParent:Render()
end


return world

end





local game
function recreate_game()
  UI_Reset()
  game = makegame()
end

local coro

function gamera()

  recreate_game()
  
  local rv
  while true do
    rv = game.tick_loop()
    if rv then break end
    coroutine.yield()
  end
  
  local dim = CreateFrame("Frame")
  dim:SetLayer(1000)
  dim:SetBackgroundColor(0.0, 0.0, 0.0, 0.9)
  dim:SetAllPoints()
  
  local results = {
    "A number that can be counted on your fingers.", -- 1
    "A number that you can count to on your own.", -- 10
    "A number that doesn't require a comma to be readable.", -- 100
    "Too many miles to drive in a single day.", -- k
    "Possibly, your yearly paycheck.", -- 10k 
    "The capacity of an old floppy disk, in bytes.", -- 100k
    "The population of a large city.", -- m
    "The number of patterns that you can make with Tangram pieces.", -- 10m
    "The number of pennies you'd need to equal the weight of a family of blue whales.", -- 100m
    "The number of people in the world.",  -- billion
    "The number of neurons in a human brain.", --10b
    "The number of stars in an average galaxy.", --100b
    "The number of fish in the sea.", --1t
    "The number of cells in the human body.", --10t
    "The number of ants on the planet.", --100t
    "The amount of oxygen in the atmosphere, in tons.", --1q
  }
  
  local top_text
  if rv[1] == "win" then
    top_text = "VICTORY!"
  else
    top_text = "Out of time."
  end
  
  local mid_text
  if rv[1] == "win" then
    mid_text = "You have made a very, very large number. One quadrillion. Congratulations."
  else
    local nmb = rv[2]
    local commaed = ""
    while #nmb > 3 do
      commaed = nmb:sub(#nmb - 2, #nmb) .. "," .. commaed
      nmb = nmb:sub(1, #nmb - 3)
    end
    if #nmb > 0 then commaed = nmb .. "," .. commaed end
    if commaed:sub(#commaed, #commaed) == "," then commaed = commaed:sub(1, #commaed - 1) end
    print(commaed)
    print(commaed[#commaed])
    
    mid_text = "Your best number: " .. commaed
  end
  
  local bottom_text
  if rv[1] == "win" then
    bottom_text = results[16]
  else
    bottom_text = results[#rv[2]]
  end
    
  local title = CreateFrame("Text", dim)
  title:SetColor(1, 1, 1)
  title:SetPoint("CENTER", UIParent, 0.5, 0, 0, 200)
  title:SetText(top_text)
  title:SetSize(90)
  
  coroutine.pause(60)
  
  local value = CreateFrame("Text", dim)
  value:SetColor(1, 1, 1)
  value:SetPoint("CENTER", UIParent, 0.5, 0, 0, 350)
  value:SetText(mid_text)
  value:SetSize(30)
  
  coroutine.pause(120)
  
  local aprox = CreateFrame("Text", dim)
  aprox:SetColor(1, 1, 1)
  aprox:SetPoint("CENTER", UIParent, 0.5, 0, 0, 500)
  aprox:SetText("Your number is approximately:")
  aprox:SetSize(45)
  
  coroutine.pause(120)
  
  local resu = CreateFrame("Text", dim)
  resu:SetColor(1, 1, 1)
  resu:SetPoint("CENTER", UIParent, 0.5, 0, 0, 600)
  resu:SetText(bottom_text)
  resu:SetSize(30)
  
  coroutine.pause(180)
  
  local sover = CreateFrame("Text", dim)
  sover:SetColor(1, 1, 1)
  sover:SetPoint("CENTER", UIParent, 0.5, 0, 0, 750)
  if rv[1] == "win" then
    sover:SetText("You won in " .. math.floor(rv[2] / 60) .. " seconds. Want to try to do better? Click to start over!")
  else
    sover:SetText("You can do better! Click to start over!")
  end
  
  while not WasKeyPressed("mouse_left") do coroutine.yield() end
  
  coro = coroutine.wrap(gamera)
  coro()
end

coro = coroutine.wrap(gamera)
coro()

function tick_loop()
  coro()
end
function render()
  game.render()
end
